#include "catch.hpp"
#include <algorithm>
#include <vector>

#include "list.hpp"


namespace {
// destructive :/
bool list_equal(list& l, const std::vector<double>& numbers) {
    // This might help to prevent hard crashes in some cases.
    // But its rather unlikely, because its the internal structure that gets people, not incrementing/decrementing sizes.
    if (l.size() != numbers.size()) {
        return false;
    }

    for (auto num : numbers) {
        if (l.front() != num) {
            return false;
        }
        l.pop_front();
    }
    return true;
}

}

TEST_CASE("Inserts, deletes and accessors") {
    list l;
    REQUIRE(l.empty());
    REQUIRE(l.size() == 0);
    SECTION("Series of push_back") {
        for (int i : {0, 2, 5, 9}) {
            l.push_back(i);
            REQUIRE(l.back() == i);
        }
        REQUIRE(l.size() == 4);
    }
    SECTION("Series of push_front") {
        for (int i : {0, 3, 7, 2}) {
            l.push_front(i);
            REQUIRE(l.front() == i);
        }
        REQUIRE(l.size() == 4);
    }
    SECTION("Series of push_back and pop_back") {
        for (int i = 0; i < 10; i += 4) {
            l.push_back(i);
            REQUIRE(l.back() == i);
            REQUIRE(l.size() == 1);
            l.pop_back();
            REQUIRE(l.size() == 0);
        }
        REQUIRE(l.size() == 0);
    }
    SECTION("Series of push_back and pop_front") {
        for (int i = 0; i < 10; i += 4) {
            l.push_back(i);
            REQUIRE(l.back() == i);
            REQUIRE(l.size() == 1);
            l.pop_front();
            REQUIRE(l.size() == 0);
        }
        REQUIRE(l.size() == 0);
    }
    SECTION("Series of push_front and pop_front") {
        for (int i = 0; i < 10; i += 4) {
            l.push_front(i);
            REQUIRE(l.back() == i);
            REQUIRE(l.size() == 1);
            l.pop_front();
            REQUIRE(l.size() == 0);
        }
        REQUIRE(l.size() == 0);
    }
    SECTION("complex 1") {
        for (int i : {0, 2, 4, 6, 8}) {
            l.push_back(i);
        }
        REQUIRE(l.size() == 5);
        l.pop_back();
        l.pop_front();
        REQUIRE(l.size() == 3);
        for (int i : {1, 3, 0}) {
            l.push_front(i);
        }
        REQUIRE(list_equal(l, { 0, 3, 1, 2, 4, 6 }));
    }
    SECTION("complex 2") {
        for (int i : {0, 2, 4}) {
            l.push_back(i);
        }
        REQUIRE(l.size() == 3);
        l.pop_back();
        l.pop_back();
        l.pop_front();
        REQUIRE(l.size() == 0);
        REQUIRE(l.empty());
        for (int i : {1, 2, 3}) {
            l.push_front(i);
            l.push_back(i);
        }
        REQUIRE(l.size() == 6);
        REQUIRE(list_equal(l, { 3, 2, 1, 1, 2, 3 }));
    }
    SECTION("complex 3") {
        for (int i : {0, 2, 4}) {
            l.push_back(i);
        }
        for (int i : {1, 3, 5}) {
            l.push_front(i);
        }
        for (int i : {1, 2}) {
            l.push_front(i);
        }
        for (int i : {3, 5}) {
            l.pop_back();
            l.push_back(i);
        }
        REQUIRE(l.size() == 8);
        REQUIRE(list_equal(l, { 2, 1, 5, 3, 1, 0, 2, 5 }));
    }
    SECTION("complex 4") {
        for (int i : {1, 2, 3, 4}) {
            l.push_back(i);
        }
        for (int i = 0; i < 2; ++i) {
            l.pop_back();
            l.pop_front();
        }
        REQUIRE(l.size() == 0);
        REQUIRE(l.empty());

        for (int i : {1, 2, 3, 4}) {
            l.push_back(i);
        }
        l.pop_back();
        l.pop_back();
        for (int i : {5, 6}) {
            l.push_back(i);
        }
        l.pop_front();
        l.pop_front();
        l.pop_front();
        REQUIRE(l.size() == 1);
        REQUIRE(!l.empty());
        for (int i : {7, 8}) {
            l.push_front(i);
        }
        REQUIRE(list_equal(l, { 8, 7, 6 }));
    }
}

TEST_CASE("empty") {
    list l;
    REQUIRE(l.empty());
    l.push_back(10);
    REQUIRE(!l.empty());
    l.pop_back();
    REQUIRE(l.empty());
    l.push_front(20);
    REQUIRE(!l.empty());
    l.pop_front();
    REQUIRE(l.empty());
    l.push_back(2);
    l.push_front(3);
    REQUIRE(!l.empty());
    l.pop_back();
    l.pop_front();
    REQUIRE(l.empty());
}

TEST_CASE("size") {
    list l;
    REQUIRE(l.size() == 0);
    l.push_back(1);
    REQUIRE(l.size() == 1);
    l.push_front(3);
    REQUIRE(l.size() == 2);
    l.push_back(3);
    l.push_back(4);
    REQUIRE(l.size() == 4);
    SECTION("pops operations should decrease size") {
        l.pop_back();
        l.pop_front();
        REQUIRE(l.size() == 2);
        l.pop_front();
        l.pop_back();
        REQUIRE(l.size() == 0);
    }
}

TEST_CASE("join") {
    list lhs, rhs;
    auto elems = { 0., 2., 4. };
    SECTION("empty lists") {
        lhs.join(rhs);
        REQUIRE(lhs.empty());
        REQUIRE(rhs.empty());
    }
    SECTION("rhs is empty") {
        for (auto e : elems) {
            lhs.push_back(e);
        }
        lhs.join(rhs);
        REQUIRE(list_equal(lhs, elems));
        REQUIRE(rhs.empty());
    }
    SECTION("lhs is empty") {
        for (auto e : elems) {
            rhs.push_back(e);
        }
        lhs.join(rhs);
        REQUIRE(list_equal(lhs, elems));
        REQUIRE(rhs.empty());
    }
    SECTION("both lists have elements in them") {
        for (auto e : elems) {
            lhs.push_back(e);
            rhs.push_back(e);
        }
        lhs.join(rhs);
        REQUIRE(list_equal(lhs, { 0, 2, 4, 0, 2, 4 }));
        REQUIRE(rhs.empty());
    }
}

TEST_CASE("reverse") {
    list l;
    SECTION("Empty list") {
        l.reverse();
    }
    SECTION("List with single element") {
        l.push_back(10);
        l.reverse();
        REQUIRE(l.back() == 10);
        SECTION("List with two elements") {
            l.push_front(20);
            l.reverse();
            REQUIRE(list_equal(l, { 10, 20 }));
        }
    }
    SECTION("List with many elements") {
        auto elems = { 0., 1., 2., 3., 4., 5., 6., 7., 8., 9. };
        for (auto e : elems) {
            l.push_back(e);
        }
        l.reverse();
        SECTION("One reverse") {
            REQUIRE(list_equal(l, { 9, 8, 7, 6, 5, 4, 3, 2, 1, 0 }));
        }
        SECTION("Two reverses cancel each other out") {
            l.reverse();
            REQUIRE(list_equal(l, elems));
        }
    }
}

TEST_CASE("unique") {
    list l;
    SECTION("Empty list") {
        l.unique();
        REQUIRE(l.size() == 0);
        REQUIRE(l.empty());
    }
    SECTION("Degenerated case 1") {
        for (int i : {1, 1, 1, 1, 1, 1, 1, 1, 1}) {
            l.push_back(i);
        }
        l.unique();
        REQUIRE(list_equal(l, { 1 }));
    }
    SECTION("Sorted elements") {
        for (int i : {1, 2, 3, 5, 5, 5}) {
            l.push_back(i);
        }
        l.unique();
        REQUIRE(list_equal(l, { 1, 2, 3, 5 }));
    }
    SECTION("Unique elements") {
        for (int i : {1, 2, 3, 4, 5}) {
            l.push_back(i);
        }
        l.unique();
        REQUIRE(list_equal(l, { 1, 2, 3, 4, 5 }));
    }
    SECTION("Unsorted elements") {
        for (int i : {1, 2, 2, 3, 3, 3, 4, 1, 2, 5, 5, 5}) {
            l.push_back(i);
        }
        l.unique();
        REQUIRE(list_equal(l, { 1, 2, 3, 4, 1, 2, 5 }));
    }
}

TEST_CASE("constructors") {
    SECTION("default constructor") {
        list l;
        REQUIRE(l.empty());
        REQUIRE(l.size() == 0);
    }
    SECTION("Vector constructor") {
        list l1(std::vector<double>{});
        REQUIRE(l1.empty());
        REQUIRE(l1.size() == 0);
        list l2({ 1 });
        REQUIRE(!l2.empty());
        REQUIRE(l2.size() == 1);
        REQUIRE(l2.front() == 1);
        REQUIRE(l2.back() == 1);
        std::vector<double> elems({ 1, 2, 3, 56, 78, 86 });
        list l3(elems);
        REQUIRE(!l3.empty());
        REQUIRE(l3.size() == 6);
        REQUIRE(list_equal(l3, elems));
    }
}

TEST_CASE("remove") {
    auto elements = std::vector<double>{ { 0, 2, 2, 3, 2, 2, 3, 3, 2, 0, 1 } };
    list l(elements);
    SECTION("Remove 2") {
        l.remove(2);
        elements.erase(std::remove(begin(elements), end(elements), 2),
                       end(elements));
        REQUIRE(l.size() == elements.size());
        REQUIRE(list_equal(l, elements));
    }
    SECTION("Remove something not in list") {
        l.remove(5);
        REQUIRE(l.size() == elements.size());
        REQUIRE(list_equal(l, elements));
    }
    SECTION("Remove head") {
        l.remove(0);
        elements.erase(std::remove(begin(elements), end(elements), 0),
                       end(elements));
        REQUIRE(l.size() == elements.size());
        REQUIRE(list_equal(l, elements));
    }
    SECTION("Remove tail") {
        l.remove(1);
        elements.erase(std::remove(begin(elements), end(elements), 1),
                       end(elements));
        REQUIRE(l.size() == elements.size());
        REQUIRE(list_equal(l, elements));
    }
    SECTION("Remove head and tail") {
        l.remove(1);
        l.remove(0);
        elements.erase(std::remove(begin(elements), end(elements), 1),
                       end(elements));
        elements.erase(std::remove(begin(elements), end(elements), 0),
                       end(elements));
        REQUIRE(l.size() == elements.size());
        REQUIRE(list_equal(l, elements));
    }
    SECTION("Remove to empty") {
        for (int i = 0; i < 4; ++i) {
            l.remove(i);
        }
        REQUIRE(l.empty());
        REQUIRE(l.size() == 0);
    }
    SECTION("Remove all") {
        list l2({ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 });
        l2.remove(1);
        REQUIRE(l2.empty());
        REQUIRE(l2.size() == 0);
    }
    SECTION("Complex 1") {
        list l2({ 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 });
        l2.remove(1);
        REQUIRE(l2.empty());
        REQUIRE(l2.size() == 0);
        SECTION("Verify that the list can still be used") {
            for (auto e : { -1, 1, 2, 3 }) {
                l2.push_back(e);
                l2.push_front(e);
            }
            REQUIRE(!l2.empty());
            REQUIRE(l2.size() == 8);
            l2.remove(3);
            REQUIRE(!l2.empty());
            REQUIRE(l2.size() == 6);
            REQUIRE(list_equal(l2, { 2, 1, -1, -1, 1, 2 }));
        }
    }
}

TEST_CASE("merge") {
    SECTION("both empty") {
        list lhs, rhs;
        lhs.merge(rhs);
        REQUIRE(rhs.empty());
        REQUIRE(rhs.size() == 0);
        REQUIRE(lhs.empty());
        REQUIRE(lhs.size() == 0);
    }
    SECTION("empty lhs") {
        list lhs;
        list rhs({0, 1, 2, 3});
        lhs.merge(rhs);
        REQUIRE(rhs.empty());
        REQUIRE(rhs.size() == 0);
        REQUIRE(!lhs.empty());
        REQUIRE(lhs.size() == 4);
        REQUIRE(list_equal(lhs, { 0, 1, 2, 3}));
    }
    SECTION("empty rhs") {
        list lhs({ 0, 1, 2, 3 });
        list rhs;
        lhs.merge(rhs);
        REQUIRE(rhs.empty());
        REQUIRE(rhs.size() == 0);
        REQUIRE(!lhs.empty());
        REQUIRE(lhs.size() == 4);
        REQUIRE(list_equal(lhs, { 0, 1, 2, 3 }));
    }
    SECTION("simple RHS append") {
        list lhs({ 0, 1, 2 });
        list rhs({ 3, 4, 5 });
        lhs.merge(rhs);
        REQUIRE(rhs.empty());
        REQUIRE(rhs.size() == 0);
        REQUIRE(!lhs.empty());
        REQUIRE(lhs.size() == 6);
        REQUIRE(list_equal(lhs, { 0, 1, 2, 3, 4, 5 }));
    }
    SECTION("simple RHS prepend") {
        list lhs({ 3, 4, 5 });
        list rhs({ 0, 1, 2 });
        lhs.merge(rhs);
        REQUIRE(rhs.empty());
        REQUIRE(rhs.size() == 0);
        REQUIRE(!lhs.empty());
        REQUIRE(lhs.size() == 6);
        REQUIRE(list_equal(lhs, { 0, 1, 2, 3, 4, 5 }));
    }
    SECTION("equal merge") {
        list lhs({ 0, 0, 0, 0 });
        list rhs({ 0, 0, 0, 0 });
        lhs.merge(rhs);
        REQUIRE(rhs.empty());
        REQUIRE(rhs.size() == 0);
        REQUIRE(!lhs.empty());
        REQUIRE(lhs.size() == 8);
        REQUIRE(list_equal(lhs, { 0, 0, 0, 0, 0, 0, 0, 0 }));
    }
    SECTION("same lists merge") {
        list lhs({0, 1, 2, 3});
        list rhs({0, 1, 2, 3});
        lhs.merge(rhs);
        REQUIRE(rhs.empty());
        REQUIRE(rhs.size() == 0);
        REQUIRE(!lhs.empty());
        REQUIRE(lhs.size() == 8);
        REQUIRE(list_equal(lhs, { 0, 0, 1, 1, 2, 2, 3, 3 }));
    }
    SECTION("Irregular merge 1") {
        list lhs({0, 1, 3, 4, 5});
        list rhs({1, 2, 2});
        lhs.merge(rhs);
        REQUIRE(rhs.empty());
        REQUIRE(rhs.size() == 0);
        REQUIRE(!lhs.empty());
        REQUIRE(lhs.size() == 8);
        REQUIRE(list_equal(lhs, { 0, 1, 1, 2, 2, 3, 4, 5 }));
    }
    SECTION("Irregular merge 2") {
        list lhs({ 1, 2, 3 });
        list rhs({-2, -2, 4, 4, 5});
        lhs.merge(rhs);
        REQUIRE(rhs.empty());
        REQUIRE(rhs.size() == 0);
        REQUIRE(!lhs.empty());
        REQUIRE(lhs.size() == 8);
        REQUIRE(list_equal(lhs, { -2, -2, 1, 2, 3, 4, 4, 5}));
    }
}

TEST_CASE("Compilation tests") {
    list l({1, 2});
    const auto& const_l = l;
    REQUIRE(const_l.front() == 1);
    REQUIRE(const_l.back() == 2);
    REQUIRE(!const_l.empty());
    REQUIRE(const_l.size() == 2);
}
